#include "m.h"

/* The SET opcode */
void OpcodeM_Set(ContextM *m)
{
   /* get the register # and advance the code pointer */
   unsigned char reg = *(m->cip++);
   /* in a more rigorous  machine we would do bounds checking */
   /* get value off bytecode stream and advance code pointer */
   m->regs[reg] = *(m->cip++);
}

/* The ADD opcode */
void OpcodeM_Add(ContextM *m)
{
   m->regs[REG_A] += m->regs[REG_B];
}

/* The MULTIPLY opcode */
void OpcodeM_Multiply(ContextM *m)
{
   m->regs[REG_A] *= m->regs[REG_B];
}

/* The END opcode */
void OpcodeM_End(ContextM *m)
{
   /* set a halt state */
   m->err = ERR_HALT;
}

/* The opcode call table */
static OPCODEM OpcodeTable[4] = 
{
   OpcodeM_Set, 
   OpcodeM_Add,
   OpcodeM_Multiply,
   OpcodeM_End
};

int Initialize_M(ContextM *m)
{
   memset(m, 0, sizeof(ContextM));
   m->oplist = OpcodeTable;
}

int Execute_M(ContextM *m, int *err)
{
   OPCODEM opfunc;
   m->cip = m->bytecode;
   m->err = 0;
   while (!m->err)
   {
      opfunc = m->oplist[*m->cip++];
      opfunc(m);
   }
   if (err)
      *err = m->err;

   /* return register A as output */
   return m->regs[REG_A];
}

