#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include "m.h"
#include "jit.h"
#include "bigtest.inc"

void dumpmem(char *data, size_t s);
unsigned int timediff(struct timeval *tm1, struct timeval *tm2)
{
	int msec = (tm2->tv_sec - tm1->tv_sec) * 1000;
	int usec1 = tm1->tv_usec / 1000;
	int usec2 = tm2->tv_usec / 1000;
	msec += (usec2-usec1);
	return msec;
}

char *dup_pcode()
{
	char *nptr;
	char *ptr;
	size_t i;
	size_t size = sizeof(test_pcode);
	nptr = (char *)malloc((size * 50) + 1);
	printf("Allocating %d size code\n", size*50);
	ptr = nptr;
	for (i=0; i<50; i++)
	{
		memcpy(ptr, test_pcode, size);
		ptr += size;
	}
	*ptr = OP_END;

	return nptr;
}

int main()
{
	ContextM m;
	MJIT jit;	
	unsigned char pcode[50];
	int err, result;
	struct timeval tm1, tm2;
	
	Initialize_M(&m);
	m.bytecode = dup_pcode();
	gettimeofday(&tm1, NULL);
	RunJIT_M(&m, &jit);
	gettimeofday(&tm2, NULL);
	printf("Codesize: %d bytes\n", jit.code_size);
	printf("JIT ran in %d milliseconds\n", timediff(&tm1, &tm2));

	gettimeofday(&tm1, NULL);
	result = ExecJIT_M(&m, &jit, &err);
	gettimeofday(&tm2, NULL);
	printf("Error: %d, Value: %d\n", err, result);
	printf("Execution time: %d milliseconds\n", timediff(&tm1, &tm2));

	gettimeofday(&tm1, NULL);
	result = Execute_M(&m, &err);
	gettimeofday(&tm2, NULL);
	printf("Error: %d, Value: %d\n", err, result);
	printf("Execution time: %d milliseconds\n", timediff(&tm1, &tm2));
}

void dumpmem(char *data, size_t s)
{
   unsigned char *p = (unsigned char *)data;
   size_t i,j;
   for (i=0; i<s; i+=16)
   {
      printf("\tdb ");
      for(j=0; j<16; j++,p++)
         printf("0%02xh%s", *p, (j==15) ? "" : ",");
      printf("\n");
   }
}

